/*
 * Copyright 2015 - 2022 gary@drinkingtea.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#pragma once

#include <ox/std/byteswap.hpp>
#include <ox/std/bstring.hpp>
#include <ox/std/memory.hpp>
#include <ox/std/string.hpp>
#include <ox/std/trace.hpp>
#include <ox/std/types.hpp>
#include <ox/std/vector.hpp>

#include "desctypes.hpp"
#include "fieldcounter.hpp"
#include "metadata.hpp"
#include "optype.hpp"
#include "typenamecatcher.hpp"
#include "types.hpp"
#include "typestore.hpp"

namespace ox {

namespace detail {

template<typename T>
static constexpr int indirectionLevels(T) noexcept {
	return 0;
}

template<typename T>
static constexpr int indirectionLevels(T *t) noexcept {
	return 1 + indirectionLevels(*t);
}

}

class TypeDescWriter {

	private:
		TypeStore *m_typeStore = nullptr;
		DescriptorType *m_type = nullptr;

	public:
		explicit constexpr TypeDescWriter(TypeStore *typeStore = nullptr) noexcept;

		template<typename T>
		constexpr Error field(const char *name, T *val, std::size_t valLen) noexcept;

		template<typename T>
		constexpr Error field(const char *name, T val) noexcept;

		template<typename T>
		constexpr Error field(const char *name, T *val) noexcept;

		template<typename T = std::nullptr_t>
		constexpr void setTypeInfo(const char *name = T::TypeName, int fields = ModelFieldCount_v<T>) noexcept;

		[[nodiscard]]
		constexpr DescriptorType *definition() noexcept {
			return m_type;
		}

		static constexpr auto opType() noexcept {
			return OpType::Reflect;
		}

	private:
		const DescriptorType *type(int8_t *val) noexcept;
		const DescriptorType *type(int16_t *val) noexcept;
		const DescriptorType *type(int32_t *val) noexcept;
		const DescriptorType *type(int64_t *val) noexcept;

		const DescriptorType *type(uint8_t *val) noexcept;
		const DescriptorType *type(uint16_t *val) noexcept;
		const DescriptorType *type(uint32_t *val) noexcept;
		const DescriptorType *type(uint64_t *val) noexcept;

		const DescriptorType *type(bool *val) noexcept;

		const DescriptorType *type(char *val) noexcept;

		const DescriptorType *type(SerStr val) noexcept;

		const DescriptorType *type(String *val) noexcept;

		template<std::size_t sz>
		constexpr const DescriptorType *type(BString<sz> *val) noexcept;

		template<typename T>
		constexpr const DescriptorType *type(T *val) noexcept;

		template<typename T>
		constexpr const DescriptorType *type(Vector<T> *val) noexcept;

		template<typename T>
		constexpr const DescriptorType *type(HashMap<String, T> *val) noexcept;

		template<typename U>
		constexpr const DescriptorType *type(UnionView<U> val) noexcept;

		const DescriptorType *getType(const String &tn, PrimitiveType t, int b) noexcept;
};

constexpr TypeDescWriter::TypeDescWriter(TypeStore *typeStore) noexcept: m_typeStore(typeStore) {
}

// array handler
template<typename T>
constexpr Error TypeDescWriter::field(const char *name, T *val, std::size_t) noexcept {
	if (m_type) {
		constexpr typename remove_pointer<decltype(val)>::type *p = nullptr;
		const auto t = type(p);
		oxAssert(t != nullptr, "field(const char *name, T *val, std::size_t): Type not found or generated");
		if (t == nullptr) {
			type(p);
		}
		m_type->fieldList.emplace_back(t, name, detail::indirectionLevels(val), t->typeName);
		return OxError(0);
	}
	return OxError(1);
}

template<typename T>
constexpr Error TypeDescWriter::field(const char *name, T val) noexcept {
	if (m_type) {
		const auto t = type(val);
		oxAssert(t != nullptr, "field(const char *name, T val): Type not found or generated");
		m_type->fieldList.emplace_back(t, name, 0, t->typeName);
		return OxError(0);
	}
	return OxError(1);
}

template<typename T>
constexpr Error TypeDescWriter::field(const char *name, T *val) noexcept {
	if (m_type) {
		const auto t = type(val);
		oxAssert(t != nullptr, "field(const char *name, T *val): Type not found or generated");
		m_type->fieldList.emplace_back(t, name, 0, t->typeName);
		return OxError(0);
	}
	return OxError(1);
}

template<std::size_t sz>
constexpr const DescriptorType *TypeDescWriter::type(BString<sz> *val) noexcept {
	return type(SerStr(val));
}

template<typename T>
constexpr const DescriptorType *TypeDescWriter::type(T *val) noexcept {
	auto [t, err] = m_typeStore->template get<T>();
	if (!err) {
		return t;
	} else {
		TypeDescWriter dw(m_typeStore);
		oxLogError(model(&dw, val));
		return dw.m_type;
	}
}

template<typename T>
constexpr const DescriptorType *TypeDescWriter::type(Vector<T> *val) noexcept {
	return type(val->data());
}

template<typename T>
constexpr const DescriptorType *TypeDescWriter::type(HashMap<String, T>*) noexcept {
	return type(static_cast<T*>(nullptr));
}

template<typename U>
constexpr const DescriptorType *TypeDescWriter::type(UnionView<U> val) noexcept {
	return type(val.get());
}

template<typename T>
constexpr void TypeDescWriter::setTypeInfo(const char *name, int) noexcept {
	m_type = m_typeStore->getInit(name);
	if constexpr(is_union_v<T>) {
		m_type->primitiveType = PrimitiveType::Union;
	} else {
		m_type->primitiveType = PrimitiveType::Struct;
	}
	m_type->preloadable = detail::preloadable<T>::value;
}

template<typename T>
Result<DescriptorType*> buildTypeDef(TypeStore *typeStore, T *val) noexcept {
	TypeDescWriter writer(typeStore);
	oxReturnError(model(&writer, val));
	return writer.definition();
}

auto writeTypeDefOC(auto *val) noexcept {
	TypeStore typeStore;
	oxRequire(def, buildTypeDef(&typeStore, val));
	return writeOC(def.get());
}

}
